#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Sprechende Uhr - Talking Clock - Deutsch
Zeitansage zu jeder Viertelminute (0, 15, 30, 45 Sekunden)
mit Hilfe von Google Text to Speech ( gTTS )

Wenn du das Programm im Terminal mit python3 zeitansage4.py startest,
dann musst du es mit im Terminal mit Strg+C beenden.

Version 1 - Funktioniert, aber überspringt Ansagen bei voller Minute
Version 2 - Ansage alle 30 Sekunden zu jeder halben und vollen Minute
Version 3 - Kommentarzeilen hinzugefügt
Aktuell Version 4 - Modifiziert für präzise Ansagen alle 15 Sekunden (0, 15, 30, 45)
und ein "time_cache" für die rechtzeitige Erzeugung der Ansagen
https://claude.ai/share/0b6d2396-2170-4b61-a04a-ee3c5167c114
Volker Lange-Janson 8. März 2025 / Modifiziert am 8. März 2025

Meinerseits darf jeder mit dem Skript machen, was er will.

Weitere Erklärungen nach dem Ende des Skripts
Das Skript benötigt Bibliotheken, sie unten im Text.
Das Skript braucht eine Internetverbindung für Google gTTS
zur Erzeugung der Sprache

Falls die tone.mp3 fehlt, kann sie mit Audacity erstellt
werden. 440 Hz Sinus, eine Sekunde lang, mono.
Die tone.mp3 im selben Ordner wie dieses Skript unterbringen.

"""

import datetime
from gtts import gTTS
from mutagen.mp3 import MP3
from playsound import playsound
import time
import os
import logging
import math

# Logging konfigurieren
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s')
logger = logging.getLogger(__name__)

# Konstanten
INTERVALS = [0, 15, 30, 45]  # Die Sekunden, zu denen die Zeit angesagt werden soll
CACHE_DIR = "time_cache"  # Verzeichnis für gecachte Zeitansagen

def setup():
    """Initialisiert das System und erstellt notwendige Verzeichnisse."""
    # Stelle sicher, dass die Tone-Datei existiert
    if not os.path.exists("tone.mp3"):
        raise FileNotFoundError("Die Datei 'tone.mp3' wurde nicht gefunden. Bitte erstellen Sie eine MP3-Datei mit einem Piepton von ca. 1 Sekunde.")
    
    # Cache-Verzeichnis für Zeitansagen erstellen
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)
        logger.info(f"Cache-Verzeichnis '{CACHE_DIR}' erstellt")

def get_next_announcement_times():
    """Berechnet die Zeiten für die nächsten Ansagen im 15-Sekunden-Takt."""
    now = datetime.datetime.now()
    result = []
    
    # Aktuelle Minute berechnen (oder nächste Minute, wenn wir kurz davor sind)
    current_minute = now.replace(second=0, microsecond=0)
    if now.second >= 45 and now.microsecond > 500000:  # Wenn wir kurz vor der nächsten Minute sind
        current_minute += datetime.timedelta(minutes=1)
    
    # Die nächsten 4 Ansagezeiten berechnen (für 0, 15, 30 und 45 Sekunden)
    for i in range(4):
        for seconds in INTERVALS:
            announcement_time = current_minute.replace(second=seconds)
            if announcement_time > now:
                result.append(announcement_time)
    
    # Nach Zeit sortieren und die nächsten 2 Ansagen zurückgeben
    result.sort()
    return result[:2]  # Wir bereiten die nächsten zwei Ansagen vor

def generate_time_str(tone_time):
    """Erzeugt den Text für die Zeitansage."""
    hour = tone_time.hour
    minute = tone_time.minute
    second = tone_time.second
    
    # Korrekte Singular/Plural-Form
    second_str = "1 Sekunde" if second == 1 else f"{second} Sekunden"
    minute_str = "1 Minute" if minute == 1 else f"{minute} Minuten"
    
    return f"Beim nächsten Ton ist es {hour} Uhr {minute_str} und {second_str}"

def get_cached_announcement(time_obj):
    """Prüft, ob eine Zeitansage bereits im Cache existiert."""
    cache_filename = f"{CACHE_DIR}/time_{time_obj.hour}_{time_obj.minute}_{time_obj.second}.mp3"
    if os.path.exists(cache_filename):
        mp3_file = MP3(cache_filename)
        return mp3_file.info.length, cache_filename
    return None, None

def create_announcement(time_obj):
    """Erstellt eine MP3-Datei für die Zeitansage und gibt die Dauer zurück."""
    time_str = generate_time_str(time_obj)
    logger.info(f"Erstelle Ansage: {time_str}")
    
    # Prüfen, ob die Ansage bereits im Cache ist
    duration, cache_file = get_cached_announcement(time_obj)
    if duration and cache_file:
        logger.info(f"Verwende gecachte Ansage: {cache_file}")
        return duration, cache_file
    
    # Neue Ansage erstellen und cachen
    cache_filename = f"{CACHE_DIR}/time_{time_obj.hour}_{time_obj.minute}_{time_obj.second}.mp3"
    audio = gTTS(text=time_str, lang='de', slow=False)
    audio.save(cache_filename)
    mp3_file = MP3(cache_filename)
    
    logger.info(f"Neue Ansage erstellt: {cache_filename} (Dauer: {mp3_file.info.length:.2f}s)")
    return mp3_file.info.length, cache_filename

def prepare_announcements(announcement_times):
    """Bereitet die Zeitansagen vor und gibt eine Liste von (Zeit, Dauer, Datei) zurück."""
    prepared = []
    for time_obj in announcement_times:
        duration, audio_file = create_announcement(time_obj)
        prepared.append((time_obj, duration, audio_file))
    return prepared

def play_announcement(announcement_time, announcement_file):
    """Spielt die Zeitansage und den Ton ab."""
    try:
        logger.info(f"Spiele Ansage ab für {announcement_time}")
        playsound(announcement_file)
        playsound("tone.mp3")
        logger.info(f"Ansage abgeschlossen für {announcement_time}")
        return True
    except Exception as e:
        logger.error(f"Fehler beim Abspielen: {e}")
        return False

def main():
    """Hauptfunktion."""
    try:
        setup()
        
        while True:
            # Die nächsten Zeitansagen vorbereiten
            next_announcements = get_next_announcement_times()
            logger.info(f"Nächste Ansagen geplant für: {[t.strftime('%H:%M:%S') for t in next_announcements]}")
            
            # Vorbereiten der Ansagen (generiert MP3-Dateien im Voraus)
            prepared_announcements = prepare_announcements(next_announcements)
            
            # Für jede vorbereitete Ansage
            for announcement_time, duration, audio_file in prepared_announcements:
                # Berechnen, wann die Ansage starten muss
                start_time = announcement_time - datetime.timedelta(seconds=duration)
                now = datetime.datetime.now()
                
                # Warten bis zur Startzeit der Ansage
                wait_seconds = (start_time - now).total_seconds()
                if wait_seconds > 0:
                    logger.info(f"Warte {wait_seconds:.2f} Sekunden bis zur Ansage für {announcement_time.strftime('%H:%M:%S')}")
                    time.sleep(wait_seconds)
                
                # Ansage abspielen
                play_announcement(announcement_time, audio_file)
                
                # Kleine Pause, um sicherzustellen, dass die nächste Berechnung nach der aktuellen Ansage beginnt
                time.sleep(1)
                
                # Nächste Ansagen neu berechnen (falls wir sehr nah an der nächsten sind)
                if datetime.datetime.now() > next_announcements[-1] - datetime.timedelta(seconds=5):
                    break
    
    except KeyboardInterrupt:
        logger.info("Programm durch Benutzer beendet")
    except Exception as e:
        logger.error(f"Unerwarteter Fehler: {e}")
        raise

if __name__ == "__main__":
    main()

"""
Entwicklung und Idee:
https://grok.com/share/bGVnYWN5_d9f89aab-dd5c-4ac7-a90d-d164682e5d76
https://claude.ai/share/0b6d2396-2170-4b61-a04a-ee3c5167c114

Zeitansage über die Soundkarte
Ansagen im 15-Sekunden-Takt (0, 15, 30, 45 Sekunden jeder Minute)

Die tone.mp3 (440Hz, 1 sec lang, mono) mit Audacity erzeugt und muss sich im selben Ordner befinden wie das Skript. 

Verbesserungen in Version 4:
- Caching der Zeitansagen zur Reduzierung der Netzwerkanfragen an Google TTS
- Präzises Timing durch vorausschauende Berechnung
- Robustes Fehlerhandling und Logging
- Garantierte Ansagen aller 15-Sekunden-Intervalle (0, 15, 30, 45)

Bibliotheken:

Um dieses Skript auszuführen, müssen Sie die folgenden externen Bibliotheken installieren:

pip install gTTS mutagen playsound
"""
